import os
import re
from importlib import import_module

from odat.defaults import DFLT_DIRS
from odat.util import ModuleNotFoundIgnore, IgnoreAllErrors, submodules_of

pjoin = os.path.join
ODAT_DIR = pjoin(os.environ.get('ODAT_DIR', DFLT_DIRS.ODAT_DIR))


def ojoin(*paths):
    return os.path.join(ODAT_DIR, *paths)


def ddir(obj):
    return [attr for attr in dir(obj) if not attr.startswith('_')]


spaces_re = re.compile('\s+')


def print_attrs(obj, with_doc_snippet=True, max_doc_length=73, sep='\n'):
    if not with_doc_snippet:
        print(*ddir(obj), sep=sep)
    else:
        def gen():
            for attr_name in ddir(obj):
                attr = getattr(obj, attr_name)
                attr_doc = (getattr(attr, '__doc__', '') or '').strip()
                if len(attr_doc) > max_doc_length:
                    attr_doc = spaces_re.sub(' ', attr_doc[:max_doc_length].strip()) + '...'
                if attr_doc:
                    yield "- " + f"{attr_name}:\n\t{attr_doc}"
                else:
                    yield "- " + attr_name

        print(*gen(), sep=sep)


if not os.path.isdir(ODAT_DIR):
    print(f"""
    Hey, you're not finished with the setup. 
    I need a rootdir to work with. 
    That was supposed to be:
        {ODAT_DIR}
    But I didn't find that directory.

    So...

    Either make that directory for me, or define one you want me to work with.

    You can specify where you want this to be by defining an environment variable called ODAT_DIR.
    That's the first place the system will look for, so will override my default    

    """)


def dacc_info_gen(on_error='ignore'):
    src = 'odat.mdat'
    for submodule_name in submodules_of(src, include_init=False):
        try:
            submodule_dotpath = src + '.' + submodule_name
            submodule = import_module(submodule_dotpath)
            d = getattr(submodule, '_meta')
            yield submodule_name, d
        except Exception as e:
            if on_error == 'raise':
                raise
            elif on_error == 'print':
                print(f"Error with {submodule_dotpath}: {e}")


daccs = dict(dacc_info_gen(on_error='ignore'))



import numpy as np
from numpy import array, hanning, fft  # TODO: Get rid of this once we use C-based spectr
from numpy.core.multiarray import zeros
from numpy.core.umath import ceil, log

# from omodel import DFLT_TILE_SIZE, DFLT_SR, DFLT_MEL_KWARGS, DFLT_MEL_MIN_AMP

DFLT_WIN_FUNC = hanning

DFLT_WF_NUM_TYPE = 'int16'
DFLT_SR = 44100
DFLT_TILE_SIZE = 2048
DFLT_HOP_SIZE = DFLT_TILE_SIZE
DFLT_TILE_STEP = DFLT_TILE_SIZE

DFLT_MEL_MIN_AMP = 1e-10
DFLT_MEL_KWARGS = dict(n_fft=2048, hop_length=512, n_mels=128)


class ParamValAssertionError(AssertionError):
    pass


class SampleRateAssertionError(ParamValAssertionError):
    pass


def tile_fft(tile, window=DFLT_WIN_FUNC, amp_function=np.abs):
    """Compute the power fft for a single tile
    """
    if callable(window):
        window = window(len(tile))
    fft_amplitudes = amp_function(np.fft.rfft(tile * window))
    return fft_amplitudes


def stft(y, n_fft=2048, hop_length=None, win_func=DFLT_WIN_FUNC):
    '''
    :param y: Mx0 audio
    :param n_fft: window size
    :param hop_length: hop size
    :return: S - DxN stft matrix
    '''

    if hop_length is None:
        hop_length = n_fft

    if win_func is not None:
        win = win_func(n_fft)
    else:
        win = 1

    # calculate STFT
    M = len(y)
    N = int(ceil(1.0 * (M - n_fft) / hop_length) + 1)  # no. windows
    S = zeros((n_fft, N), dtype='complex')
    for f in range(N - 1):
        S[:, f] = y[f * hop_length:n_fft + f * hop_length] * win
    x_end = y[(N - 1) * hop_length:]
    S[:len(x_end), N - 1] = x_end
    S[:, N - 1] *= win
    S = fft.fft(S, axis=0)
    S = S[:n_fft // 2 + 1, :]

    return S


def get_wf_to_spectr_func(tile_size=DFLT_TILE_SIZE, tile_step=None, win_func=DFLT_WIN_FUNC):
    """
    Get a spectr_of_wf function that computes spectrograms (lists of spectrums really) for given tile_size and tile_step

    :param tile_size: n_fft, also called "window size" or "buffer size"
    :param tile_step: By default, tile_size (i.e. non-overlapping and gapless sfft window support
    :return: a spectr_of_wf function that computes spectrograms
    >>> from random import randint
    >>>
    >>> tile_size = 2048
    >>> n_tiles = 21
    >>> spectr_of_wf = get_wf_to_spectr_func(tile_size)
    >>>
    >>> wf = [randint(-30000, 30000) for x in range(n_tiles * tile_size)]
    >>> spectr_of_wf(wf).shape
    (1025, 21)
    >>>
    >>> wf = [randint(-30000, 30000) for x in range(n_tiles * tile_size - 1)]
    >>> spectr_of_wf(wf).shape
    (1025, 20)
    >>>
    >>> wf = [randint(-30000, 30000) for x in range(n_tiles * tile_size + 1)]
    >>> spectr_of_wf(wf).shape
    (1025, 21)
    >>>
    >>> tile_step = 1024
    >>> spectr_of_wf = get_wf_to_spectr_func(tile_size, tile_step=tile_step)
    >>>
    >>> wf = [randint(-30000, 30000) for x in range(n_tiles * tile_size)]
    >>> spectr_of_wf(wf).shape
    (1025, 41)
    >>> spectr_of_wf = get_wf_to_spectr_func()
    >>> spectra = spectr_of_wf([0.] * 2048 * 3)
    >>> list(sum(spectra))
    [0.0, 0.0, 0.0]
    >>> all([x == 0 for x in sum(spectra)])
    True
    """

    if tile_step is None:
        tile_step = tile_size

    def spectr_of_wf(wf):
        n_tiles = max(1 + (len(wf) - tile_size) // tile_step, 0)
        return spectr(array(wf), sr=None, tile_size=tile_size, tile_step=tile_step)[:n_tiles, :].T

    return spectr_of_wf


def wf_to_spectr_func(tile_size=DFLT_TILE_SIZE, tile_step=None, win_func=None):
    """Like get_wf_to_spectr_func, but with frequencies in columns"""
    _wf_to_spectr = get_wf_to_spectr_func(tile_size, tile_step, win_func=win_func)

    def wf_to_spectr(wf):
        return _wf_to_spectr(wf).T

    return wf_to_spectr


dflt_wf_to_spectr = wf_to_spectr_func()


class Spectr(object):
    def __init__(self, tile_size=DFLT_TILE_SIZE, tile_step=None, win_func=DFLT_WIN_FUNC, **lin_trans_kwargs):
        self.tile_size = tile_size
        if tile_step is None:
            tile_step = tile_size
        self.tile_step = tile_step

    def __call__(self, wf):
        n_tiles = max(1 + (len(wf) - self.tile_size) // self.tile_step, 0)
        return spectr(array(wf), sr=None, tile_size=self.tile_size, tile_step=self.tile_step)[:n_tiles, :].T


def abs_stft(wf, tile_size=DFLT_TILE_SIZE, tile_step=None, win_func=DFLT_WIN_FUNC):
    if len(wf) > 0:
        return abs(stft(wf, n_fft=tile_size, hop_length=tile_step, win_func=win_func)).T
    else:
        return np.array([])


def spectr(wf, sr=None, tile_size=DFLT_TILE_SIZE, tile_step=None, win_func=DFLT_WIN_FUNC):
    tile_step = tile_step or tile_size
    return abs_stft(wf, tile_size=tile_size, tile_step=tile_step, win_func=win_func)


# Note: Removing until librosa gets it's act together
# try:
#     import librosa
#
#
#     def log_mel_of_wf_sr(wf, sr, common_sr=DFLT_SR):
#         if common_sr is None:
#             feat_mat = librosa.feature.melspectrogram(wf, sr=sr, **DFLT_MEL_KWARGS)
#         else:
#             _assert_sample_rate_if_not_none(common_sr, sr)
#             feat_mat = librosa.feature.melspectrogram(wf, sr=common_sr, **DFLT_MEL_KWARGS)
#         return librosa.db_to_amplitude(feat_mat, ref=DFLT_MEL_MIN_AMP).T  # TODO: Look into
#
#
#     def onset_strength_and_spectral_contrast(wf, sr):
#         onset_strength = librosa.onset.onset_strength(wf, sr)
#         spectral_contrast = librosa.feature.spectral_contrast(wf, sr)
#         return np.vstack((onset_strength, spectral_contrast)).T  # TODO: Look into
#
#
#     # TODO: Look into this!
#     def intensity_and_spectral_features(wf, sr, tile_size=DFLT_TILE_SIZE, tile_step=None):
#         tile_step = tile_step or tile_size
#         S = abs_stft(wf, tile_size=tile_size, tile_step=tile_step)
#         log_intensity = log(S.sum(axis=0) + 1)
#         onset_strength = librosa.onset.onset_strength(sr=sr, S=S)
#         spectral_centroid = librosa.feature.spectral_centroid(sr=sr, S=S, n_fft=tile_size, hop_length=tile_step)
#         spectral_contrast = librosa.feature.spectral_contrast(sr=sr, S=S, n_fft=tile_size, hop_length=tile_step)
#         return np.vstack((log_intensity, onset_strength, spectral_centroid, spectral_contrast)).T  # TODO: Look into
# except ImportError as e:
#     from warnings import warn
#
#     warn(str(e))


def intensity_and_normalized_spectr(wf, sr=None, tile_size=DFLT_TILE_SIZE,
                                    tile_step=None, intensity_normalization_factor=1):
    tile_step = tile_step or tile_size
    S = abs_stft(wf, tile_size=tile_size, tile_step=tile_step)
    intensity = S.sum(axis=0)
    S /= intensity
    return np.vstack((log(intensity + 1) / intensity_normalization_factor, S))


def intensity_normalized_spectr(wf, sr=None, tile_size=DFLT_TILE_SIZE, tile_step=None):
    tile_step = tile_step or tile_size
    S = abs_stft(wf, tile_size=tile_size, tile_step=tile_step)
    intensity = S.sum(axis=0)
    S /= intensity
    return S


def log_spectr(wf, sr=None, tile_size=DFLT_TILE_SIZE, tile_step=None):
    return log(1 + spectr(wf, sr=sr, tile_size=tile_size, tile_step=tile_step))


def _assert_sample_rate_if_not_none(assert_sr, sr=None):
    if sr is not None and sr != assert_sr:
        raise SampleRateAssertionError("Sample rate was {}: Should be {}".format(sr, assert_sr))



import os
import importlib_resources  # replace when moving to 3.9+


class ModuleNotFoundIgnore:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is ModuleNotFoundError:
            pass
        return True


class IgnoreAllErrors:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        return True


def is_submodule_path(path):
    path = str(path)
    return path.endswith('.py')


def module_name(path):
    name, ext = os.path.splitext(os.path.basename(path))
    return name


def submodules_of(pkg, include_init=True):
    f = importlib_resources.files(pkg)
    g = map(module_name, filter(is_submodule_path, f.iterdir()))
    if include_init:
        return g
    else:
        return filter(lambda name: name != '__init__', g)


import os

######################### FILE PATHS AND STORES INITIALIZATION #########################
# TODO: Move spec to config so it's editable
dflt_path_templates = (
    "{name}",  # dataname IS the path
    "~/odat/{name}",  # can be found in ~/odat
    "~/Downloads/{name}",  # can be found in Downloads
    "~/{name}",  # can be found in home folder
)
normalize_path = lambda p: os.path.abspath(os.path.expanduser(p))


def find_datapath(name, path_templates=dflt_path_templates):
    path_options = map(lambda x: normalize_path(x.format(name=name)), path_templates)
    r = next(filter(os.path.exists, path_options), None)
    if r is None:
        raise ValueError(f"Did not find a file named {name} in any of these folders: {dflt_path_templates}")
    else:
        return r



import os
from itertools import chain
from collections import Counter
from warnings import warn
import re
from io import BytesIO

import pandas as pd
import numpy as np
import soundfile as sf

from py2store.stores.s3_store import S3BinaryStore

from py2store import (
    LocalJsonStore,
    add_ipython_key_completions,
    mk_read_only,
    lazyprop,
)
from odat import ODAT_DIR, ojoin, pjoin

odat_subdir = 'iatis'
dflt_data_rootdir = ojoin(odat_subdir)

from py2store import myconfigs

ro_client_kwargs = None
try:
    ro_client_kwargs = myconfigs['oto.ini']['s3_all_ro']
except Exception:
    pass

if ro_client_kwargs is None:
    warn("You don't have the myconfigs['oto.ini']['s3_all_ro'], which I need to be able to access s3. "
         "This means you won't be able to access the audio data.")


def delete_empty_records(s):
    """To delete all empty (i.e. length 0) items of a store"""
    for k, v in s.items():
        if len(v) == 0:
            del s[k]


def mk_subdict_extractor(fields):
    """Makes a function that generates field-inclusion defined subdicts from a dict iterator."""

    def extract(dict_list):
        for dd in dict_list:
            yield {k: dd[k] for k in fields}

    return extract


from py2store import KvReader, kv_wrap


# Pattern: Explicit store pattern (dict wrapping store). Todo: Use relevant tools
class DfGroupReader(KvReader):
    def __init__(self, df, by=None, **groupby_kwargs):
        self.src = df.groupby(by, **groupby_kwargs)

    def __iter__(self):
        yield from self.src

    def __contains__(self, k):
        return k in self.src

    def __len__(self):
        return len(self.src)

    def __getitem__(self, k):
        return self.src.get_group(k)


@kv_wrap.outcoming_vals(lambda x: x["_id"].values.tolist())
class TagSrefs(DfGroupReader):
    pass


s3_path_p = re.compile('^(?P<protocol>[^:]+)://(?P<bucket>[^/]+)/(?P<path>.*)')
s3_chks_tmpl = 's3://{group}/learn_mode/sounds/{user}/{channel}/{day}/{bt}_{tt}'
local_chks_tmpl = '{group}/{user}/{channel}/{day}/{bt}_{tt}'


def parse_sref(k):
    m = s3_path_p.match(k)
    if m:
        protocol, bucket, path = m.groups()
        return protocol, bucket, path
    else:
        raise KeyError(f"Your key couldn't be parsed: {k}")


def mk_sref_bytes_reader(resource_kwargs):
    """Make a function that takes s3://... template keys to s3 resrouces, and returns bytes"""

    def get_bytes_from_s3(k):
        protocol, bucket, path = parse_sref(k)
        assert protocol == 's3', f"That wasn't an s3 path: {k}"
        b = S3BinaryStore(bucket, '', resource_kwargs)[path]
        return b

    return get_bytes_from_s3


def mk_sref_wf_reader(resource_kwargs):
    get_bytes_from_s3 = mk_sref_bytes_reader(resource_kwargs)

    def get_wfsr(k):
        return sf.read(BytesIO(get_bytes_from_s3(k)))

    return get_wfsr


get_bytes_from_s3 = mk_sref_bytes_reader(ro_client_kwargs)
get_wfsr = mk_sref_wf_reader(ro_client_kwargs)


class Dacc:
    _fv_mgc_subdir = 'iatis_fv_mgc'
    _fv_mgc_subpath = '{group}/{user}'
    _sref_tag_df_pkl = 'sref_tag_df.pkl'
    cache_sref_tag_df = True

    def __init__(self, data_rootdir=dflt_data_rootdir):
        self.data_rootdir = data_rootdir

    def djoin(self, *paths):
        return os.path.join(self.data_rootdir, *paths)

    def random_wfsrs_tagged(self, tag, n=1):
        srefs = self.sref_tag_store[tag]
        n_srefs = len(srefs)
        srefs_selected = np.random.choice(srefs, n, replace=n_srefs < n)
        for sref in srefs_selected:
            yield sref, get_wfsr(sref)

    def wf_tag_gen(self, tags, assert_sr):
        """Get a (wf, tag) pairs iterator for given tags"""
        for sref, tag, (wf, sr) in self.sref_tag_wfsr_gen(tags):
            if assert_sr is not None:
                assert sr == assert_sr, f"You are asserting sr={assert_sr}, but I got {sr}"
            yield wf, tag

    def tag_wf_gen(self, tags, assert_sr):
        yield from ((tag, wf) for wf, tag in self.wf_tag_gen(tags, assert_sr))

    def sref_tag_bytes_gen(self, tags):
        for tag in tags:
            for sref in self.sref_tag_store[tag]:
                yield sref, tag, get_bytes_from_s3(sref)

    def sref_tag_wfsr_gen(self, tags):
        for tag in tags:
            for sref in self.sref_tag_store[tag]:
                yield sref, tag, get_wfsr(sref)

    @lazyprop
    def sref_tag_store(self):
        return TagSrefs(self.sref_tag_df, 'tag')

    # Pattern: Caching. TODO: Use local file py2store caching for this
    @lazyprop
    def sref_tag_df(self):
        """df containing tag and _id (s3 key) columns.
        (Took 1mn to make on my computer)
        """
        if os.path.isfile(self._path_sref_tag_df_pkl):
            df = pd.read_pickle(self._path_sref_tag_df_pkl)
        else:
            extract_tag_and_id = mk_subdict_extractor(('_id', 'tag'))
            df = pd.DataFrame(list(chain.from_iterable(map(extract_tag_and_id,
                                                           self.fv_mgc_store.values()))))
            if self.cache_sref_tag_df:
                df.to_pickle(self._path_sref_tag_df_pkl)

        return df

    @lazyprop
    def _fv_mgc_zip_filepath(self):
        return self.djoin(self._fv_mgc_subdir + '.zip')

    @lazyprop
    def _path_sref_tag_df_pkl(self):
        return self.djoin(self._sref_tag_df_pkl)

    @lazyprop
    def _path_fv_mgc_path_format(self):
        return self.djoin(self._fv_mgc_subdir,
                          self._fv_mgc_subpath)

    @lazyprop
    def fv_mgc_store(self):
        return mk_read_only(LocalJsonStore(self._path_fv_mgc_path_format))

    @lazyprop
    def tag_counts(self):
        return pd.Series(Counter(self.sref_tag_df['tag'])).sort_values(ascending=False)

    @lazyprop
    def tagged_fvs_of_users(self):
        """A store whose keys are (group, user) pairs and values are (_id, deviceHwId, tag, fv) DataFrames"""
        from py2store import FilesOfZip, kv_wrap
        import json

        class TaggedFvsTrans:
            def _id_of_key(self, k):
                return self.rootdir + '/'.join(k)

            def _key_of_id(self, _id):
                return tuple(_id[len(self.rootdir):].split('/'))

            def _obj_of_data(self, data):
                return pd.DataFrame(json.loads(data.decode()))

            # The key mapping depends on a rootdir attr that needs to be defined in the class that will be wrapped.
            # To indicate this to the user (and your IDE linter) you can do this:

            # rootdir = None  # to be defined in subclass
            # or this:
            @property
            def rootdir(self):
                raise NotImplementedError("This is supposed to be set in the class that will be wrapped")

        from py2store import kv_wrap

        @kv_wrap(TaggedFvsTrans)
        class TaggedFvs(FilesOfZip):
            rootdir = self._fv_mgc_subdir + '/'  # you could also write a new __init__ that would take it as a arg (+ the ones of FilesOfZip)

        return TaggedFvs(self._fv_mgc_zip_filepath)


_meta = dict(
    name='iatis',
    description='Over 5 million tagged sounds',
    mk_dacc=Dacc
)



import os
from functools import partial
from py2store import myconfigs
from py2store.ext.audio import WavLocalFileStore
from slang.core import KvDataSource
import re
import numpy as np

from odat.utils.chunkers import fixed_step_chunker

DFLT_CHUNKER = partial(fixed_step_chunker, chk_size=2048)

config_filename = 'xe.json'
DFLT_LOCAL_SOURCE_DIR = myconfigs.get_config_value(config_filename, 'local_source_dir')

p_first_path_component = re.compile('[^/]+')


def first_path_component(x):
    m = p_first_path_component.match(x)
    if m:
        return m.group(0)


def mk_kv_data_source(local_source_dir=DFLT_LOCAL_SOURCE_DIR,
                      key_to_tag=first_path_component, key_filt=None):
    assert local_source_dir is not None, "You need to specify the directory where your data is"
    s = WavLocalFileStore(os.path.join(local_source_dir, '{tag}/{filename}.wav'))
    return KvDataSource(s, key_to_tag=key_to_tag, key_filt=key_filt)


def mk_dacc(local_source_dir=DFLT_LOCAL_SOURCE_DIR,
            key_to_tag=first_path_component,
            wf_to_chk=DFLT_CHUNKER,
            key_filt=None):
    assert local_source_dir is not None, "You need to specify the directory where your data is"
    s = WavLocalFileStore(os.path.join(local_source_dir, '{tag}/{filename}.wav'))

    class Dacc(KvDataSource):
        def chk_tag_gen(self):
            for k, tag, chk in self.key_tag_chks_gen(wf_to_chk=DFLT_CHUNKER):
                yield np.array(chk), tag

        def wf_tag_gen(self):
            for k, tag, wf in self.key_tag_wf_gen():
                yield wf, tag

    return Dacc(s, key_to_tag=key_to_tag, key_filt=key_filt)


_meta = dict(
    name='xe',
    description='Fridge compressor data',
    mk_dacc=mk_dacc
)



from functools import partial
from collections import deque
from itertools import dropwhile, islice, chain
from io import BytesIO
import os
import numpy as np
import pandas as pd
import soundfile as sf
from py2store import FilesOfZip, KvReader, wrap_kvs, filt_iter, cached_keys
from odat.utils.chunkers import clever_chunker
from odat.util import find_datapath

DFLT_CHK_SIZE = 2048
DFLT_CHK_STEP = None
DFLT_CHUNKER = partial(clever_chunker, tile_size=DFLT_CHK_SIZE, tile_step=DFLT_CHK_STEP)
# A function to be applied to the raw fft output
DFLT_SPECTRA_TRANS = lambda x: np.abs(x)
dataname = 'ss1.zip'
DFLT_WF_FOLDER = 'wav'

DFLT_CONTEXT_FILENAME = 'Context_all.csv'

DFLT_ZIP_FILE_PATH = find_datapath(dataname)


######################################### STORES ###########################################
def mk_wf_store(zip_file_path=DFLT_ZIP_FILE_PATH, wf_folder=DFLT_WF_FOLDER):
    """
    Make a store accessing the wav files in the zip
    :param zip_file_path: str, path to the zip file
    :param wf_folder: str, the name of the folder containing the wav files within the zip file
    :return: a store with key being the name of the wav files and values the array of samples
    """
    wf_folder_size = len(wf_folder)

    @wrap_kvs(
        key_of_id=lambda x: x[wf_folder_size + 1:],
        id_of_key=lambda x: os.path.join(wf_folder, x),
        obj_of_data=lambda b: sf.read(BytesIO(b), dtype='int16')[0])
    @filt_iter(filt=lambda x: x.endswith('.wav'))
    class WfStore(FilesOfZip):
        """Waveform access. Keys are .wav filenames and values are numpy arrays of int16 waveform."""
        pass

    # making the two stores on which the dacc will be build
    wfs = WfStore(zip_file_path)  # access the wf
    # wfs_list = list(wfs)
    # t = ", ".join(list(wfs)[:3])
    return wfs


def mk_context_store(zip_file_path=DFLT_ZIP_FILE_PATH):
    """
    Make a store accessing the csv files in the zip
    :param zip_file_path: str, path to the zip file
    :return: a store with key being the name of the csv files and values the panda dataframe of the
             files
    """

    @filt_iter(filt=lambda x: x.endswith('.csv'))
    @wrap_kvs(obj_of_data=lambda b: pd.read_csv(BytesIO(b)))
    class CSVStore(FilesOfZip):
        pass

    context_store = CSVStore(zip_file_path)
    return context_store


class AnnotStore(KvReader):
    def __init__(self, annots_df, key_cols=None):
        self.annots_df = annots_df.set_index(key_cols)
        self.annots_df['filename'] = self.annots_df.index

    def __iter__(self):
        yield from self.annots_df.index.values

    def __getitem__(self, k):
        return self.annots_df.loc[k]

    def __len__(self):
        return len(self.annots_df)

    def __contains__(self, k):
        return k in self.annots_df.index


def mk_annots_store(zip_file_path=DFLT_ZIP_FILE_PATH, context_name=DFLT_CONTEXT_FILENAME):
    context_store = mk_context_store(zip_file_path=zip_file_path)
    return AnnotStore(context_store[context_name], key_cols='filename')


####################################### DATA ACCESS #######################################
class Dacc:
    """The Dacc is an object that will allow access to both the wf store and the annotation store
    and let us get chunks, spectra and fvs. Again, don't spend time trying to understand everything
    """

    def __init__(self,
                 wfs=mk_wf_store,
                 annots=mk_annots_store,
                 annot_to_tag=lambda x: x.truth,
                 extra_info=lambda x: x.to_dict(),
                 chunker=DFLT_CHUNKER,
                 spectra_trans=DFLT_SPECTRA_TRANS):
        if callable(wfs):  # if callable, consider it a factory, and make the annots
            wfs = wfs()
        if callable(annots):  # if callable, consider it a factory, and make the annots
            annots = annots()
        self.wfs = wfs
        self.annots = annots
        self.annot_to_tag = annot_to_tag
        if extra_info is None:
            self.extra_info = lambda x: None
        else:
            self.extra_info = extra_info
        self.chunker = chunker
        self.spectra_trans = spectra_trans

    # this allows you to create an instance of the dacc class where the data is filtered out
    # you could also filter out externally, but this can be convenient
    @classmethod
    def with_key_filt(cls, key_filt, wfs, annots, annot_to_tag, extra_info, chunker, spectra_trans):
        filtered_annots = cached_keys(annots, keys_cache=key_filt)
        filtered_wfs = cached_keys(wfs, keys_cache=key_filt)
        return cls(filtered_wfs, filtered_annots, annot_to_tag, extra_info, chunker, spectra_trans)

    def _get_tile_fft(self, tile):
        fft_amplitudes = self.spectra_trans(np.fft.rfft(tile))
        return fft_amplitudes

    def wf_tag_gen(self):
        """Yields the (wf, tag) pairs"""
        for k in self.annots:
            try:
                wf = self.wfs[k]
                annot = self.annots[k]
                yield wf, self.annot_to_tag(annot)
            except KeyError:
                pass

    def chk_tag_gen(self):
        """Yields the (chunk, tag) pairs"""
        for wf, tag, extra in self.wf_tag_gen():
            for chk in self.chunker(wf):
                yield chk, tag, extra

    def fft_tag_gen(self):
        """Yields the (fft, tag) pairs"""
        for chk, tag, extra in self.chk_tag_gen():
            yield self._get_tile_fft(chk), tag, extra


_meta = dict(
    name='ss',
    description='Loose screws in rotating car dashboard',
    mk_dacc=Dacc
)



from itertools import islice


def fixed_step_chunker(it, chk_size, chk_step=None, start_at=None, stop_at=None, return_tail=False):
    """
      a function to get (an iterator of) segments (bt, tt) of chunks from an iterator (or list)
      of the for [it_1, it_2...], given a chk_size, chk_step, and a start_at and a stop_at.
      The start_at, stop_at act like slices indices for a list: start_at is included and stop_at is excluded

      :param it: iterator of elements of any type
      :param chk_size: length of the chunks
      :param chk_step: step between chunks
      :param start_at: index of the first term of the iterator at which we begin building the chunks (inclusive)
      :param stop_at: index of the last term from the iterator included in the chunks
      :param return_tail: if set to false, only the chunks with max element with index less than stop_at are yielded
      if set to true, any chunks with minimum index value no more than stop_at are returned but they contain
      term with index no more than stop_at
      :return: an iterator of the chunks

      1) If stop_at is not None and return_tail is False:
         will return all full chunks with maximum element index less than stop_at
         or until the iterator is exhausted. Only full chunks are returned here.

      2) If stop_at is not None and return_tail is True:
         will return all full chunks as above along with possibly cut off chunks
         containing one term whose index is stop_at-1 or one (last) term which is the
         last element of it

      3) If stop_at is None and return_tail is False:
         will return all full chunks with maximum element index less or equal to the last
         element of it

      4) If stop_at is None and return_tail is True:
         will return all full chunks with maximum element index less or equal to the last
         element of it plus cut off chunks whose maximum term index is the last term of it

        # testing chk_step < chk_size with return_tail=TRUE, stop and start_at PRESENT
        # and stop_at SMALLER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=1, start_at=2, stop_at=5, return_tail=True)
        >>> it = list(range(1, 17, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[3, 4, 5], [4, 5], [5]]

        # testing chk_step < chk_size with return_tail=FALSE, stop and start_at PRESENT
        # and stop_at SMALLER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=1, start_at=2, stop_at=5, return_tail=False)
        >>> it = list(range(1, 17, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[3, 4, 5]]

        # testing chk_step < chk_size with return_tail=TRUE, stop and start_at PRESENT
        # and stop_at LARGER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=1, start_at=1, stop_at=20, return_tail=True)
        >>> it = list(range(1, 17, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[2, 3, 4], [3, 4, 5], [4, 5, 6], [5, 6, 7], [6, 7, 8], [7, 8, 9], [8, 9, 10], [9, 10, 11], [10, 11, 12], [11, 12, 13], [12, 13, 14], [13, 14, 15], [14, 15, 16], [15, 16], [16]]

        # testing chk_step < chk_size with return_tail=FALSE, stop and start_at PRESENT
        # and stop_at LARGER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=1, start_at=1, stop_at=20, return_tail=False)
        >>> it = list(range(1, 17, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[2, 3, 4], [3, 4, 5], [4, 5, 6], [5, 6, 7], [6, 7, 8], [7, 8, 9], [8, 9, 10], [9, 10, 11], [10, 11, 12], [11, 12, 13], [12, 13, 14], [13, 14, 15], [14, 15, 16]]

        # testing chk_step = chk_size with return_tail=TRUE, stop and start_at PRESENT
        # and stop_at SMALLER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=3, start_at=1, stop_at=7, return_tail=True)
        >>> it = list(range(1, 17, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[2, 3, 4], [5, 6, 7]]

        # testing chk_size > len(it) with return_tail=False, no stop_at or start_at
        >>> f = lambda it: fixed_step_chunker(it, chk_size=30, chk_step=3, start_at=None, stop_at=None, return_tail=False)
        >>> it = list(range(1, 17, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        []

        # testing chk_size > len(it) with return_tail=True, no stop_at or start_at
        >>> f = lambda it: fixed_step_chunker(it, chk_size=30, chk_step=3, start_at=None, stop_at=None, return_tail=True)
        >>> it = list(range(1, 17, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16], [4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16], [7, 8, 9, 10, 11, 12, 13, 14, 15, 16], [10, 11, 12, 13, 14, 15, 16], [13, 14, 15, 16], [16]]

        # testing chk_step > chk_size with return_tail=TRUE, stop and start_at PRESENT
        # and stop_at SMALLER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=4, start_at=1, stop_at=7, return_tail=True)
        >>> it = list(range(1, 17, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[2, 3, 4], [6, 7]]

        # testing chk_step > chk_size with return_tail=FALSE, stop and start_at PRESENT
        # and stop_at SMALLER than the largest index of it
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=4, start_at=1, stop_at=7, return_tail=False)
        >>> it = list(range(1, 17, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[2, 3, 4]]

        # testing chk_step > chk_size with return_tail=FALSE, stop and start_at NOT PRESENT
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=4, start_at=None, stop_at=None, return_tail=False)
        >>> it = list(range(1, 17, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[1, 2, 3], [5, 6, 7], [9, 10, 11], [13, 14, 15]]

        # testing chk_step > chk_size with return_tail=TRUE, stop and start_at NOT PRESENT
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=4, start_at=None, stop_at=None, return_tail=True)
        >>> it = list(range(1, 19, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[1, 2, 3], [5, 6, 7], [9, 10, 11], [13, 14, 15], [17, 18]]

        # testing chk_step > chk_size with return_tail=TRUE, stop and start_at NOT PRESENT
        # with negative values in the iterator
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=4, start_at=None, stop_at=None, return_tail=True)
        >>> it = list(range(-10, 19, 1))
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [[-10, -9, -8], [-6, -5, -4], [-2, -1, 0], [2, 3, 4], [6, 7, 8], [10, 11, 12], [14, 15, 16], [18]]

        # testing chk_step > chk_size with return_tail=TRUE, stop and start_at NOT PRESENT
        # with items of various types in the iterator
        >>> f = lambda it: fixed_step_chunker(it, chk_size=3, chk_step=2, start_at=None, stop_at=None, return_tail=True)
        >>> it = ['a', 3, -10, 9.2, str, [1,2,3], set([10,20])]
        >>> A = list(f(it)); B = list(f(iter(it)));  # trying the function on it (a list) and iter(it) (and iterator)
        >>> assert A == B  # it and iter(it) should give the same thing!
        >>> A  # and that thing is:
        [['a', 3, -10], [-10, 9.2, <type 'str'>], [<type 'str'>, [1, 2, 3], set([10, 20])], [set([10, 20])]]
       """

    if chk_step is None:
        chk_step = chk_size

    if start_at is None:
        start_at = 0

    # if the input is a list
    if hasattr(it, '__getslice__'):

        if stop_at is None:
            stop_at = len(it)
        else:
            stop_at = min(len(it), stop_at)

        it = it[start_at:stop_at]
        it_minus_chk_length = len(it) - chk_size
        n_full_chk_to_return = max(int((it_minus_chk_length / chk_step) + 1), 0)
        bt = 0
        tt = bt + chk_size

        for i in range(n_full_chk_to_return):
            yield it[bt:tt]
            bt += chk_step
            tt += chk_step

        if return_tail:
            while len(it[bt:tt]) > 0:
                yield it[bt:tt]
                bt += chk_step
                tt += chk_step

    # if the input is an iterator
    else:
        if chk_step < chk_size:

            it = islice(it, start_at, stop_at)
            chk = list(islice(it, chk_size))

            while len(chk) == chk_size:
                yield chk
                chk = chk[chk_step:] + list(islice(it, chk_step))

        else:
            it = islice(it, start_at, stop_at)
            chk = list(islice(it, chk_size))
            gap = chk_step - chk_size

            while len(chk) == chk_size:
                yield chk
                chk = list(islice(it, gap, gap + chk_size))

        if return_tail:
            while len(chk) > 0:
                yield chk
                chk = chk[chk_step:]


from collections import deque
from itertools import dropwhile, chain


def clever_chunker(gen_to_chunk, tile_size, tile_step=None):
    """
    Yield the items from gen_to_chunk into groups of consecutive terms of size tile_size.
    Tile_step controls the offsets from one group to another.
    :param gen_to_chunk: a generator/list
    :param tile_size: the size of the chunks
    :param tile_step: the step from one chunk to the next
    :return: a generator
    """
    if tile_step is None:
        tile_step = tile_size
    # add or remove terms to make sure the one liner initialize correctly
    init = (tile_size - int(tile_size / tile_step) * tile_step)
    if init > 0 and tile_step > init:
        gen_to_chunk = chain(['junk'] * (tile_step - init), gen_to_chunk)
        d = deque([], tile_size)
    else:
        d = deque(list(islice(gen_to_chunk, init)), tile_size)
    # the one liner
    res = dropwhile(lambda x: len(x) < tile_size,
                    map(lambda x: tuple(d), map(d.extend, zip(*([iter(gen_to_chunk)] * tile_step)))))
    return res



import os


class DFLT_DIRS:
    ODAT_DIR = '~/odat'


# for a in filter(lambda x: not x.startswith('_'), dir(DFLT_DIRS)):
for k, v in DFLT_DIRS.__dict__.items():
    if not k.startswith('_'):
        setattr(DFLT_DIRS, k, os.path.expanduser(v))



from py2store import FilesOfZip, filt_iter, wrap_kvs, lazyprop
import soundfile as sf
from io import BytesIO
import re
from dataclasses import dataclass
from typing import Callable

import numpy as np

from odat.utils.spectro import wf_to_spectr_func, DFLT_WIN_FUNC
from odat.util import find_datapath

dataname = 'conveyor_states.zip'
ASSERT_SR = 44100
DFLT_ZIP_FILE_PATH = find_datapath(dataname)

def get_wf(wav_bytes, assert_sr=ASSERT_SR):
    wf, sr = sf.read(BytesIO(wav_bytes))
    if assert_sr is not None:
        assert sr == assert_sr, f"Sample rate was {sr}. I expected {assert_sr}"
    return wf


key_to_tag_pattern = re.compile('(\w+).wav$')


def key_to_tag(key):
    tag, *_ = key_to_tag_pattern.search(key).groups()
    return tag


@wrap_kvs(obj_of_data=get_wf)
@filt_iter(filt=lambda x: not x.startswith('__MACOSX') and not x.endswith('DS_Store'))
class Wavs(FilesOfZip):
    pass


dflt_wf_to_spectr = wf_to_spectr_func(
    tile_size=2048, tile_step=None, win_func=DFLT_WIN_FUNC)


@dataclass
class Dacc:
    zip_filepath: str = DFLT_ZIP_FILE_PATH
    key_to_tag: Callable = key_to_tag
    wf_to_spectr: Callable = dflt_wf_to_spectr

    @lazyprop
    def wfs(self):
        return Wavs(self.zip_filepath)

    def wf_tag_gen(self):
        for k, wf in self.wfs.items():
            tag = self.key_to_tag(k)
            yield wf, tag

    def tag_spectr_gen(self):
        for k, wf in self.wfs.items():
            tag = self.key_to_tag(k)
            for spectra in self.wf_to_spectr(wf):
                yield tag, spectra

    @lazyprop
    def _spectra_tags(self):
        tags, spectras = list(map(np.array, zip(*self.tag_spectr_gen())))
        return spectras, tags

    @property
    def spectras(self):
        return self._spectra_tags[0]

    @property
    def tags(self):
        return self._spectra_tags[1]

    def get_xy_sample(self, n=400):
        idx = np.random.choice(range(len(self.tags)), size=n, replace=False)
        return self.spectras[idx], self.tags[idx]


_meta = dict(
    name='conveyor_belts_01.zip',
    description='Conveyor belts',
    mk_dacc=Dacc,
)
